/*****************************************************************************/
/*
                                DynDNSrpt.c

This is a CGI-based script, suitable for WASD, VMS Apache, OSU, and (perhaps)
Purveyor, to report on the DynDNSupd application.  Just drop it into your
server's CGI environment.  It is the complementary component to DYNDNSUPD.C
and reports using the logical names and log file created by it.

The easiest way to allow it access to the DYNDNSUPD.C created log file is to
INSTALL it with SYSPRV.  Also requires WORLD to access the DynDNSupd process
data.  To obtain

  $ INSTALL REPLACE CGI-BIN:[000000]DYNDNSRPT.EXE /PRIV=(SYSPRV,WORLD)

(or modified for your server configuration).

To provide some measure of access control the script insists on being subject
to server authorization by checking that CGI variable REMOTE_USER is not empty. 
You will need to configure this aspect using whatever your Web server requires.

Requires a browser that supports Cascading Style Sheets and JavaScript
(although it will operate minimally without it).

The [filter] button (default) displays all significant (formatted) log events
(startup, date changes, errors, updates, etc.  Button [all] displays all
(formatted) events.  The [raw] button the non-formatted log file contents.


LOGICAL NAMES
-------------
DYNDNSRPT_PARAM      equivalent to (overrides) the command line
                     parameters/qualifiers (define as a system-wide logical)

DYNDNSRPT$DBUG       turns on all "if (Debug)" statements


BUILD DETAILS
-------------
See BUILD_DYNDNSRPT.COM procedure.


COPYRIGHT
---------
Copyright (C) 2006 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY (update SOFTWAREVN as well!)
---------------
29-JUL-2006  MGD  v1.0.0, initial
*/
/*****************************************************************************/

#define SOFTWAREVN "1.0.0"
#define SOFTWARE_COPYRIGHT "Copyright (C) 2006 Mark G.Daniel - GPL licensed"
#define SOFTWARENM "DYNDNSRPT"
#ifdef __ALPHA
#  define SOFTWAREID SOFTWARENM " AXP-" SOFTWAREVN
#endif
#ifdef __ia64
#  define SOFTWAREID SOFTWARENM " IA64-" SOFTWAREVN
#endif
#ifdef __VAX
#  define SOFTWAREID SOFTWARENM " VAX-" SOFTWAREVN
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unixio.h>

/* VMS-related header files */
#include <descrip.h>
#include <jpidef.h>
#include <libdef.h>
#include <libdtdef.h>
#include <lib$routines.h>
#include <lnmdef.h>
#include <psldef.h>
#include <starlet.h>
#include <ssdef.h>
#include <stsdef.h>
#include <syidef.h>

/* application related header file */
#include <cgilib.h>

#define BOOL int
#define TRUE 1
#define FALSE 0

#define LOGNAME_AUTH         "DYNDNSUPD_AUTH"
#define LOGNAME_HOST_CHECK   "DYNDNSUPD_HOST_CHECK"
#define LOGNAME_HOST_UPDATE  "DYNDNSUPD_HOST_UPDATE"
#define LOGNAME_HOST         "DYNDNSUPD_HOST"
#define LOGNAME_IP           "DYNDNSUPD_IP"
#define LOGNAME_LOG          "DYNDNSUPD_LOG"
#define LOGNAME_NOW          "DYNDNSUPD_NOW"
#define LOGNAME_PARAM        "DYNDNSUPD_PARAM"
#define LOGNAME_PID          "DYNDNSUPD_PID"
#define LOGNAME_RETURN       "DYNDNSUPD_RETURN"
#define LOGNAME_TIME_LAST    "DYNDNSUPD_TIME_LAST"
#define LOGNAME_TIME_NEXT    "DYNDNSUPD_TIME_NEXT"
#define LOGNAME_TIME_UPDATE  "DYNDNSUPD_TIME_UPDATE"

#define DYNDNS_HOME     "http://www.dyndns.com/"
#define DYNDNS_ACCOUNT  "https://www.dyndns.com/account/"

#ifndef __VAX
#   pragma nomember_alignment
#endif

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) (!((x) & STS$M_SUCCESS))
 
#define FI_LI __FILE__, __LINE__

/******************/
/* global storage */
/******************/

BOOL  Debug;

char  DynDNSip [32],
      LogFileName [256],
      MyHostName [256],
      SoftwareID [96],
      TimeLast [24],
      TimeNext [24],
      TimeUpdate [24],
      UpdateReturn [128],
      Utility [] = "DYNDNSRPT";


/**************/
/* JavaScript */
/**************/

char  JavaScript [] = "\
function doOnLoad () {\n\
   bottomLog();\n\
}\n\
function toggleLog (btn) {\n\
  if (btn == \"Refresh\")\n\
     window.location.reload();\n\
  else\n\
  if (btn == \"Raw\")\n\
     window.location.replace(scriptName+\"?log=raw\");\n\
  else\n\
  if (btn == \"All\")\n\
     window.location.replace(scriptName+\"?log=all\");\n\
  else\n\
     window.location.replace(scriptName);\n\
  return false;\n\
}\n\
function bottomLog () {\n\
   objLog = document.getElementById(\"idLog\");\n\
   objLog.scrollTop = objLog.scrollHeight;\n\
}\n\
";

/***************/
/* style sheet */
/***************/

char  StyleSheet [] = "\
body { \n\
   padding: 1em;\n\
   padding-top: 0.5em;\n\
   color: #000000;\n\
   background-color: #ffffff;\n\
}\n\
th { \n\
   white-space: nowrap;\n\
   text-align: right;\n\
   vertical-align: top;\n\
   padding-right: 0.3em;\n\
}\n\
td { \n\
   text-align: left;\n\
   vertical-align: top;\n\
}\n\
td.logNoWrap { \n\
   vertical-align: top;\n\
   white-space: nowrap;\n\
   padding-left: 0.5em;\n\
   padding-right: 0.5em;\n\
}\n\
td.logNoWrapRight { \n\
   text-align: right;\n\
   vertical-align: top;\n\
   white-space: nowrap;\n\
   padding-left: 0.5em;\n\
   padding-right: 0.5em;\n\
}\n\
td.logWrap { \n\
   vertical-align: top;\n\
   padding-left: 0.5em;\n\
   padding-right: 0.5em;\n\
}\n\
input {\n\
   border: 1px solid #666666;\n\
   background-color: #eeeeee;\n\
   width: 6em;\n\
   padding: 0.2em;\n\
   margin-right: 0.3em;\n\
}\n\
div.logText {\n\
   height: 15em;\n\
   overflow: auto;\n\
   border: 1px solid #666666;\n\
   background-color: #eeeeee;\n\
   padding: 0.5em;\n\
   font-family: monospace;\n\
   white-space: pre;\n\
}\n\
";

/**************/
/* prototypes */
/**************/

char* DynDnsUpdProcess ();
void GetParameters ();
int ReadFileIntoMemory (char*, char**, int*);
void ReportLogFile ();
char* SysTrnLnm (char*, char*);
char* SyiNodeName ();
char* TimeWhen (char*);
BOOL strsame (char*, char*, int);
int strzcpy (char*, char*, int);

/*****************************************************************************/
/*
'argc/argv' are only required to support OSU, in particular CgiLibOsuInit().
*/

main
(
int argc,
char *argv[]
)
{
   int  UpdateErrorCount,
        UpdateFatalCount,
        UpdateSuccessCount,
        UpdateWarningCount;
   char  *cptr, *sptr, *zptr,
         *CgiScriptNamePtr;
   char  CurrentTimeStr [24];
   $DESCRIPTOR (CurrentTimeStrDsc, CurrentTimeStr);
   $DESCRIPTOR (CurrentTimeFaoDsc, "!20%D\0");

   /*********/
   /* begin */
   /*********/

   sprintf (SoftwareID, "%s (%s)", SOFTWAREID, CgiLibEnvironmentVersion());

   sys$fao (&CurrentTimeFaoDsc, 0, &CurrentTimeStrDsc, 0);

   if (Debug = (getenv ("DYNDNSRPT$DBUG") != NULL))
      fprintf (stdout, "Content-type: text/plain\r\n\r\n");
   CgiLibEnvironmentSetDebug (Debug);

   GetParameters ();

   CgiLibEnvironmentInit (argc, argv, FALSE);

   cptr = CgiLibVar ("REMOTE_USER");
   if (!*cptr)
   {
      CgiLibResponseError (FI_LI, 0,
         "Script must be subject to authorization.");
      exit (SS$_NORMAL);
   }

   cptr = SysTrnLnm (LOGNAME_HOST, NULL);
   if (cptr)
      strzcpy (MyHostName, cptr, sizeof(MyHostName));
   else
      strcpy (MyHostName, "?");

   cptr = SysTrnLnm (LOGNAME_IP, NULL);
   if (cptr)
      strzcpy (DynDNSip, cptr, sizeof(DynDNSip));
   else
      strcpy (TimeLast, "?");

   cptr = SysTrnLnm (LOGNAME_TIME_LAST, NULL);
   if (cptr)
      strzcpy (TimeLast, cptr, sizeof(TimeLast));
   else
      strcpy (TimeLast, "?");

   cptr = SysTrnLnm (LOGNAME_TIME_NEXT, NULL);
   if (cptr)
      strzcpy (TimeNext, cptr, sizeof(TimeNext));
   else
      strcpy (TimeNext, "?");

   cptr = SysTrnLnm (LOGNAME_TIME_UPDATE, NULL);
   if (cptr)
      strzcpy (TimeUpdate, cptr, sizeof(TimeUpdate));
   else
      strcpy (TimeUpdate, "?");

   cptr = SysTrnLnm (LOGNAME_RETURN, NULL);
   if (cptr)
   {
      strzcpy (UpdateReturn, cptr, sizeof(UpdateReturn));
      sscanf (UpdateReturn, "SUCCESS:%d WARNING:%d ERROR:%d FATAL:%d",
              &UpdateSuccessCount, &UpdateWarningCount,
              &UpdateErrorCount, &UpdateFatalCount);
   }
   else
   {
      strcpy (UpdateReturn, "?");
      UpdateSuccessCount = UpdateWarningCount =
         UpdateErrorCount = UpdateFatalCount = 0;
   }

   cptr = SysTrnLnm (LOGNAME_LOG, NULL);
   if (cptr)
      strzcpy (LogFileName, cptr, sizeof(LogFileName));
   else
      LogFileName[0] = '\0';

   CgiScriptNamePtr = CgiLibVar ("SCRIPT_NAME");

   CgiLibResponseHeader (200, "text/html");

   fprintf (stdout,
"<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"copyright\" CONTENT=\"%s\">\n\
<META NAME=\"environment\" CONTENT=\"%s\">\n\
<STYLE TYPE=\"text/css\">\n\
<!--\n\
%s\
-->\n\
</STYLE>\n\
<SCRIPT LANGUAGE=\"JavaScript\">\n\
<!--\n\
var scriptName = \"%s\"\n\
%s\
-->\n\
</SCRIPT>\n\
<TITLE>DynDNSupd for %s</TITLE>\n\
</HEAD>\n\
<BODY ONLOAD=\"doOnLoad()\">\n\
<FORM ACTION=\"%s\">\n\
<TABLE CELLPADDING=\"0\" CELLSPACING=\"7\" BORDER=\"0\">\n\
<TR><TD></TD>\
<TH STYLE=\"font-size:110%%;text-align:left;text-decoration:underline;\">\
DynDNSupd for %s</TH></TR>\n\
<TR><TD></TD><TD><A HREF=\"%s\" TARGET=\"_blank\">%s</A>\
<BR><A HREF=\"%s\" TARGET=\"_blank\">%s</A></TD></TR>\n\
<TR><TH>Process:</TH><TD>%s</TD></TR>\n\
<TR><TH>Host Name%s:</TH><TD>%s</TD></TR>\n\
<TR><TH>Current IP:</TH><TD>%s</TD></TR>\n\
<TR><TH>Current Time:</TH><TD>%s</TD></TR>\n\
<TR><TH>Last Update:</TH><TD>%s &nbsp;&ndash;&nbsp; %s<BR>\n\
<TABLE CELLPADDING=\"0\" CELLSPACING=\"1\" BORDER=\"0\">\n\
<TR><TH>Success:</TH><TD>%d</TD></TR>\n\
<TR><TH>Warning:</TH><TD>%d</TD></TR>\n\
<TR><TH>Error:</TH><TD>%d</TD></TR>\n\
<TR><TH>Fatal:</TH><TD>%d</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
<TR><TH>Last Check:</TH><TD>%s &nbsp;&ndash;&nbsp; %s</TD></TR>\n\
<TR><TH>Next Check:</TH><TD>%s &nbsp;&ndash;&nbsp; %s</TD></TR>\n\
<TR><TH>Log File:</TH><TD>%s &nbsp;&ndash;&nbsp; ",
            SoftwareID, SOFTWARE_COPYRIGHT, CgiLibEnvironmentName(),
            StyleSheet,
            CgiScriptNamePtr, JavaScript,
            SyiNodeName(),
            CgiScriptNamePtr,
            SyiNodeName(),
            DYNDNS_HOME, DYNDNS_HOME,
            DYNDNS_ACCOUNT, DYNDNS_ACCOUNT,
            DynDnsUpdProcess(),
            strchr(MyHostName,',') ? "s" : "", MyHostName,
            DynDNSip, CurrentTimeStr,
            TimeUpdate, TimeWhen(TimeUpdate),
            UpdateSuccessCount, UpdateWarningCount,
            UpdateErrorCount, UpdateFatalCount,
            TimeLast, TimeWhen(TimeLast),
            TimeNext, TimeWhen(TimeNext),
            LogFileName);

   if (LogFileName[0])
      ReportLogFile ();
   else
      fprintf (stdout, "</TD></TR>\n");

   fprintf (stdout,
"</TABLE>\n\
</FORM>\n\
</BODY>\n</HTML>\n");

   exit (SS$_NORMAL);
}

/*****************************************************************************/
/*
Provide the log file in a scrollable window as raw or formatted.
*/

void ReportLogFile ()

{
   BOOL  LogAll,
         LogRaw;
   int  status,
        LogTextLength;
   float  LogMBytes;
   char  CategoryChar;
   char  *cptr, *elptr, *lptr, *sptr, *zptr,
         *LogTextPtr;
   char  DateStr [12],
         ErrorStr [128],
         KeyWord [32],
         PrevDateStr [12],
         TimeStr [12];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ReportLogFile()\n");

   LogAll = LogRaw = FALSE;
   cptr = CgiLibVar ("FORM_LOG");
   if (tolower(*cptr) == 'a')
      LogAll = TRUE;
   else
   if (tolower(*cptr) == 'r')
      LogRaw = TRUE;

   status = ReadFileIntoMemory (LogFileName, &LogTextPtr, &LogTextLength);
   if (VMSnok (status))
   {
      sprintf (LogTextPtr = ErrorStr, "ERROR READING LOG: %%X%08.08X", status);
      LogTextLength = 0;
      LogRaw = TRUE;
      LogAll = FALSE;
   }

   if (LogTextLength < 1000)
      fprintf (stdout, "%d Bytes", LogTextLength);
   else
   if (LogTextLength < 1048576)
      fprintf (stdout, "%d kBytes", LogTextLength/1024);
   else
   {
      LogMBytes = (float)LogTextLength / 1048576.0;
      fprintf (stdout, "%.3f MBytes", LogMBytes);
   }
   fprintf (stdout, "</TD></TR>\n");

   fprintf (stdout, "<TR><TD></TD><TD><DIV ID=\"idLog\" CLASS=\"logText\">");

   if (LogRaw)
      fputs ((char*)CgiLibHtmlEscape(LogTextPtr,LogTextLength,NULL,0), stdout);
   else
   {
      fprintf (stdout,
"<TABLE CELLPADDING=\"0\" CELLSPACING=\"1\" BORDER=\"0\">"); 

      lptr = LogTextPtr;
      while (*lptr)
      {
         for (cptr = lptr; *cptr && *cptr != '\n'; cptr++);
         elptr = cptr;
         if (*elptr) *elptr++ = '\0';
         while (*elptr && *elptr == '\n') elptr++;

         if (*lptr) CategoryChar = *lptr++;
         while (*lptr == ' ') lptr++;

         /* not interested in the source code line number */
         while (*lptr && *lptr != ' ')lptr++;
         while (*lptr == ' ') lptr++;

         zptr = (sptr = DateStr) + sizeof(DateStr)-1;
         while (*lptr && *lptr != ' ' && sptr < zptr) *sptr++ = *lptr++;
         *sptr = '\0';
         while (*lptr == ' ') lptr++;
         if (strcmp (DateStr, PrevDateStr))
            strcpy (PrevDateStr, DateStr);
         else
            DateStr[0] = '\0';

         if (!LogAll)
         {
            if (!DateStr[0] && CategoryChar == 'I')
            {
               lptr = elptr;
               continue;
            }
         }

         zptr = (sptr = TimeStr) + sizeof(TimeStr)-1;
         while (*lptr && *lptr != ' ' && sptr < zptr) *sptr++ = *lptr++;
         *sptr = '\0';
         while (*lptr == ' ') lptr++;

         zptr = (sptr = KeyWord) + sizeof(KeyWord)-1;
         while (*lptr && *lptr != ' ' && sptr < zptr) *sptr++ = *lptr++;
         *sptr = '\0';

         while (*lptr == ' ') lptr++;

         if (CategoryChar == 'E' || CategoryChar == 'F')
            cptr = " STYLE=\"background-color:#ff3333\"";
         else
         if (CategoryChar == 'W')
            cptr = " STYLE=\"background-color:#ffff33\"";
         else
         if (CategoryChar == 'U')
            cptr = " STYLE=\"background-color:#00ff00\"";
         else
            cptr = "";

         fprintf (stdout,
"<TR>\
<TD CLASS=\"logNoWrapRight\">%s</TD>\
<TD CLASS=\"logNoWrap\">%s</TD>\
<TD CLASS=\"logNoWrap\"%s>%s</TD>\
<TD CLASS=\"logWrap\">%s</TD>\
</TR>\n",
                  (char*)CgiLibHtmlEscape(DateStr,-1,NULL,0),
                  (char*)CgiLibHtmlEscape(TimeStr,-1,NULL,0),
                  cptr, (char*)CgiLibHtmlEscape(KeyWord,-1,NULL,0),
                  (char*)CgiLibHtmlEscape(lptr,-1,NULL,0));

         lptr = elptr;
      }

      fprintf (stdout, "</TABLE>");
   }

   fprintf (stdout,
"</SPAN></TD></TR>\n\
<TR><TD></TD><TD>\
<INPUT TYPE=\"submit\" NAME=\"log\" VALUE=\"Refresh\" \
ONCLICK=\"return toggleLog(this.value)\"> \
<INPUT TYPE=\"submit\" NAME=\"log\" VALUE=\"Filter\" %s\
ONCLICK=\"return toggleLog(this.value)\"> \
<INPUT TYPE=\"submit\" NAME=\"log\" VALUE=\"All\" %s\
ONCLICK=\"return toggleLog(this.value)\"> \
<INPUT TYPE=\"submit\" NAME=\"log\" VALUE=\"Raw\" %s\
ONCLICK=\"return toggleLog(this.value)\"> \
</TD></TR>\n",
            !LogAll && !LogRaw ? "STYLE=\"font-weight:bold;\" " : "",
            LogAll ? "STYLE=\"font-weight:bold;\" " : "",
            LogRaw ? "STYLE=\"font-weight:bold;\" " : "");
}

/****************************************************************************/
/*
Read the file contents specified by 'FileName' into memory, set the pointer
at 'FileTextPtr' to the contents and the file size at 'FileSizePtr'.  Returns a
VMS status value that should be checked.
*/ 

int ReadFileIntoMemory
(
char *FileName,
char **FileTextPtr,
int *FileSizePtr
)
{
   int  status,
        Bytes,
        BytesRemaining,
        BufferCount;
   char  *cptr,
         *BufferPtr,
         *LinePtr;
   FILE  *FilePtr;
   stat_t  FstatBuffer;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ReadFileIntoMemory() |%s|\n", FileName);

   if (FileTextPtr != NULL) *FileTextPtr = NULL;
   if (FileSizePtr != NULL) *FileSizePtr = 0;

   FilePtr = fopen (FileName, "r", "ctx=rec","shr=put");
   if (FilePtr == NULL)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "fopen() %%X%08.08X\n", status);
      return (status);
   }

   if (fstat (fileno(FilePtr), &FstatBuffer) < 0)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "fstat() %%X%08.08X\n", status);
      fclose (FilePtr);
      return (status);
   }

   Bytes = FstatBuffer.st_size;
   if (Debug) fprintf (stdout, "%d bytes\n", Bytes);
   /* a little margin for error ;^) */
   Bytes += 32;

   BufferPtr = calloc (Bytes, 1);
   if (BufferPtr == NULL)
   {
      CgiLibResponseError (FI_LI, vaxc$errno, "calloc()");
      exit (SS$_NORMAL);
   }
   BytesRemaining = Bytes;
   LinePtr = BufferPtr;

   BufferCount = 0;
   while (fgets (LinePtr, BytesRemaining, FilePtr) != NULL)
   {
      /** if (Debug) fprintf (stdout, "|%s|\n", LinePtr); **/
      for (cptr = LinePtr; *cptr && *cptr != '\r'; cptr++);
      if (*(unsigned short*)cptr == '\r\n')
      {
         *(unsigned short*)cptr = '\n\0';
         cptr++;
      }
      BufferCount += cptr - LinePtr;
      BytesRemaining -= cptr - LinePtr;
      LinePtr = cptr;
   }
   fclose (FilePtr);

   if (Debug) fprintf (stdout, "%d |%s|\n", BufferCount, BufferPtr);

   if (FileTextPtr != NULL) *FileTextPtr = BufferPtr;
   if (FileSizePtr != NULL) *FileSizePtr = BufferCount;

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Given a date/time in standard string format calculate how far that is in
days/hours/minutes from current time and return a pointer to a string
describing that.
*/

char* TimeWhen (char *TimeStr)

{
   static unsigned long  LibDeltaDays = LIB$K_DELTA_DAYS,
                         LibDeltaHours = LIB$K_DELTA_HOURS,
                         LibDeltaMinutes = LIB$K_DELTA_MINUTES;
   static  CurrentBinTime [2];

   int  len, status;
   unsigned long  DeltaDays,
                  DeltaHours,
                  DeltaMinutes;
   unsigned long  ResultBinTime [2],
                  WhenBinTime [2];
   char  *cptr,
         *AgoPtr,
         *InPtr;
   char  WhenStr [128];
   $DESCRIPTOR (TimeStrDsc, TimeStr);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TimeWhen() |%s|\n", TimeStr);

   if (!CurrentBinTime[0]) sys$gettim (&CurrentBinTime);

   TimeStrDsc.dsc$w_length = strlen(TimeStr);
   status = lib$convert_date_string (&TimeStrDsc, &WhenBinTime,
                                     0, 0, 0, 0);
   if (VMSnok (status)) return ("ERROR lib$convert_date_string()");

   status = lib$sub_times (&CurrentBinTime, &WhenBinTime, &ResultBinTime);
   if (status == LIB$_NEGTIM)
   {
      AgoPtr = "";
      InPtr = "in ";
      status = lib$sub_times (&WhenBinTime, &CurrentBinTime, &ResultBinTime);
   }
   else
   {
      AgoPtr = " ago";
      InPtr = "";
   }
   if (VMSnok (status)) return ("ERROR lib$sub_times()");

   status = lib$cvt_from_internal_time (&LibDeltaDays, &DeltaDays,
                                        &ResultBinTime);
   if (VMSnok (status)) return ("ERROR lib$cvt_from_internal_time(days)");

   status = lib$cvt_from_internal_time (&LibDeltaHours, &DeltaHours,
                                        &ResultBinTime);
   if (VMSnok (status)) return ("ERROR lib$cvt_from_internal_time(hours)");

   status = lib$cvt_from_internal_time (&LibDeltaMinutes, &DeltaMinutes,
                                        &ResultBinTime);
   if (VMSnok (status)) return ("ERROR lib$cvt_from_internal_time(minutes)");

   DeltaMinutes -= DeltaHours * 60;
   if (*InPtr) DeltaMinutes++;
   DeltaHours -= DeltaDays * 24;

   if (DeltaDays && DeltaHours)
      len = sprintf (WhenStr, "%s%d day%s %d hour%s%s",
                     InPtr, DeltaDays, DeltaDays == 1 ? "" : "s",
                     DeltaHours, DeltaHours == 1 ? "" : "s", AgoPtr);
   else
   if (DeltaDays)
      len = sprintf (WhenStr, "%s%d day%s%s",
                     InPtr, DeltaDays, DeltaDays == 1 ? "" : "s", AgoPtr);
   else
   if (DeltaHours)
      len = sprintf (WhenStr, "%s%d hour%s%s",
                     InPtr, DeltaHours, DeltaHours == 1 ? "" : "s", AgoPtr);
   else
      len = sprintf (WhenStr, "%s%d minute%s%s",
                     InPtr, DeltaMinutes, DeltaMinutes == 1 ? "" : "s", AgoPtr);

   cptr = calloc (1, len);
   if (!cptr) exit (vaxc$errno);
   strcpy (cptr, WhenStr);

   return (cptr);
}

/*****************************************************************************/
/*
Get the detached update process data and return a pointer to a string
containing a summary.  Because it uses a PID from a logical which may be stale
(if the deatched process has long exited for some reason) if the process exists
check what image it's executing and if it's not DYNDNSUPD.EXE then just
consider the PID old information.
*/

char* DynDnsUpdProcess ()

{
   static char  ProcessStr [96];

   static char  *StateNameArray [] = {
                "1234","COLPG","MWAIT","CEF","PFW","LEF","LEFO",
                "HIB","HIBO","SUSP","SUSPO","FPG","COM","COMO","CUR" };

   static unsigned long  GetJpiControlFlags = JPI$M_IGNORE_TARGET_STATUS;

   static unsigned long  JpiPid,
                         JpiState;
   static char  JpiImagName [256],
                JpiPrcNam [16];

   static struct
   {
      unsigned short  buf_len;
      unsigned short  item;
      unsigned char  *buf_addr;
      unsigned short  *short_ret_len;
   }
      JpiItems [] =
   {
      { sizeof(GetJpiControlFlags), JPI$_GETJPI_CONTROL_FLAGS,
        (void*)&GetJpiControlFlags, 0 },
      { sizeof(JpiPid), JPI$_PID, (void*)&JpiPid, 0 },
      { sizeof(JpiImagName), JPI$_IMAGNAME, (void*)&JpiImagName, 0 },
      { sizeof(JpiPrcNam), JPI$_PRCNAM, (void*)&JpiPrcNam, 0 },
      { sizeof(JpiState), JPI$_STATE, (void*)&JpiState, 0 },
      { 0,0,0,0 }
   };

   int  status;
   unsigned long  ProcessContext;
   char  *cptr, *sptr;
   struct {
      unsigned short  iosb$w_status;
      unsigned short  iosb$w_bcnt;
      unsigned long  iosb$l_reserved;
   } IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DynDnsUpdProcess()\n");

   cptr = SysTrnLnm (LOGNAME_PID, NULL);
   if (!cptr)
   {
      status = SS$_NOLOGNAM;
      sprintf (ProcessStr, "%%X%08.08X &nbsp;(no PID logical)", status);
      return (ProcessStr);
   }
   sscanf (cptr, "%X", &ProcessContext);
   if (Debug) fprintf (stdout, "ProcessContext %08.08X\n", ProcessContext);

   status = sys$getjpiw (0, &ProcessContext, 0, &JpiItems, &IOsb, 0, 0);
   if (VMSok (status)) status = IOsb.iosb$w_status;
   if (VMSnok (status))
   {
      if (status == SS$_NOPRIV)
         sprintf (ProcessStr, "%%X%08.08X &nbsp;(no privilege)", status);
      else
      if (status == SS$_NONEXPR)
         sprintf (ProcessStr, "%%X%08.08X &nbsp;(nonexistent process)", status);
      else
         sprintf (ProcessStr, "%%X%08.08X", status);
      return (ProcessStr);
   }

   for (cptr = JpiImagName; *cptr && *cptr != ';'; cptr++);
   if (*cptr == ';') *cptr-- = '\0';
   while (cptr > JpiImagName && *cptr != ']') cptr--;
   if (*cptr == ']') cptr++;
   if (Debug) fprintf (stdout, "|%s|\n", cptr);

   /* if not executing the update image then it's not our target */
   if (strcmp (cptr, "DYNDNSUPD.EXE"))
   {
      status = SS$_NONEXPR;
      sprintf (ProcessStr, "%%X%08.08X &nbsp;(image not found)", status);
      return (ProcessStr);
   }

   JpiPrcNam[15] = '\0';
   for (cptr = JpiPrcNam; *cptr && *cptr != ' '; cptr++);
   *cptr = '\0';

   if (JpiState > 0 && JpiState <= 14)
      sptr = StateNameArray[JpiState];
   else
      sprintf (sptr = StateNameArray[0], "%04.04X", JpiState);

   sprintf (ProcessStr, "pid:%08.08X &nbsp;prcnam:%s &nbsp;state:%s",
            JpiPid, JpiPrcNam, sptr);
   return (ProcessStr);
}

/*****************************************************************************/
/*
Return a pointer to the node name of this system.
*/

char* SyiNodeName ()

{
   static unsigned short  NodeNameLen;
   static char  NodeName [16];

   static struct
   {
      unsigned short  buf_len;
      unsigned short  item;
      unsigned char  *buf_addr;
      void  *ret_len;
   }
   SyiItems [] =
   {
     { sizeof(NodeName)-1, SYI$_NODENAME, (void*)NodeName, &NodeNameLen },
     { 0,0,0,0 }
   };

   int  status;

   /*********/
   /* begin */
   /*********/

   if (NodeName[0]) return (NodeName);
   status = sys$getsyi (0, 0, 0, &SyiItems, 0, 0, 0);
   if (VMSnok (status)) exit (status);
   NodeName[NodeNameLen] = '\0';
   return (NodeName);
}

/*****************************************************************************/
/*
Translate an exec mode  logical name using LNM$FILE_DEV.  Return a pointer to
the value string, or NULL if the name does not exist.  If 'LogValue' is
supplied the logical name is translated into that (assumed to be large enough),
otherwise it's translated into an internal static buffer.
*/

char* SysTrnLnm
(
char *LogName,
char *LogValue
)
{
   static unsigned char  ExecMode = PSL$C_EXEC;
   static unsigned short  ShortLength;
   static char  StaticLogValue [256];
   static $DESCRIPTOR (LogNameDsc, "");
   static $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");
   static struct {
      short int  buf_len;
      short int  item;
      void  *buf_addr;
      unsigned short  *ret_len;
   } LnmItems [] =
   {
      { 255, LNM$_STRING, 0, &ShortLength },
      { 0,0,0,0 }
   };

   int  status;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SysTrnLnm() |%s|\n", LogName);

   LogNameDsc.dsc$a_pointer = LogName;
   LogNameDsc.dsc$w_length = strlen(LogName);
   if (LogValue)
      cptr = LnmItems[0].buf_addr = LogValue;
   else
      cptr = LnmItems[0].buf_addr = StaticLogValue;

   status = sys$trnlnm (0, &LnmFileDevDsc, &LogNameDsc, &ExecMode, &LnmItems);
   if (Debug) fprintf (stdout, "sys$trnlnm() %%X%08.08X\n", status);
   if (!(status & 1))
   {
      if (Debug) fprintf (stdout, "|(null)|\n");
      return (NULL);
   }

   cptr[ShortLength] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", cptr);
   return (cptr);
}

/*****************************************************************************/
/*
Get "command-line" parameters, whether from the command-line or from a
configuration symbol or logical containing the equivalent.  OSU scripts have
the 'method', 'url' and 'protocol' supplied as P1, P2, P3 (these being detected
and used by CGILIB), and are of no interest to this function.
*/

void GetParameters ()

{
   static char  CommandLine [256];
   static unsigned long  Flags = 0;

   int  status,
        SkipParameters;
   unsigned short  Length;
   char  ch;
   char  *aptr, *cptr, *clptr, *sptr;
   $DESCRIPTOR (CommandLineDsc, CommandLine);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetParameters()\n");

   if ((clptr = getenv ("DYNDNSRPT_PARAM")) == NULL)
   {
      /* get the entire command line following the verb */
      if (VMSnok (status =
          lib$get_foreign (&CommandLineDsc, 0, &Length, &Flags)))
         exit (status);
      (clptr = CommandLine)[Length] = '\0';
   }

   /* if OSU environment then skip P1, P2, P3 */
   if (getenv ("WWWEXEC_RUNDOWN_STRING") != NULL)
      SkipParameters = 3;
   else
      SkipParameters = 0;

   aptr = NULL;
   ch = *clptr;
   for (;;)
   {
      if (aptr != NULL && *aptr == '/') *aptr = '\0';
      if (!ch) break;

      *clptr = ch;
      if (Debug) fprintf (stdout, "clptr |%s|\n", clptr);
      while (*clptr && isspace(*clptr)) *clptr++ = '\0';
      aptr = clptr;
      if (*clptr == '/') clptr++;
      while (*clptr && !isspace (*clptr) && *clptr != '/')
      {
         if (*clptr != '\"')
         {
            clptr++;
            continue;
         }
         cptr = clptr;
         clptr++;
         while (*clptr)
         {
            if (*clptr == '\"')
               if (*(clptr+1) == '\"')
                  clptr++;
               else
                  break;
            *cptr++ = *clptr++;
         }
         *cptr = '\0';
         if (*clptr) clptr++;
      }
      ch = *clptr;
      if (*clptr) *clptr = '\0';
      if (Debug) fprintf (stdout, "aptr |%s|\n", aptr);
      if (!*aptr) continue;

      if (SkipParameters)
      {
         SkipParameters--;
         continue;
      }

      if (strsame (aptr, "/DBUG", -1))
      {
         Debug = TRUE;
         continue;
      }

      if (*aptr == '/')
      {
         fprintf (stdout, "%%%s-E-IVQUAL, unrecognized qualifier\n \\%s\\\n",
                  Utility, aptr+1);
         exit (STS$K_ERROR | STS$M_INHIB_MSG);
      }

      fprintf (stdout, "%%%s-E-MAXPARM, too many parameters\n \\%s\\\n",
               Utility, aptr);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }
}

/****************************************************************************/
/*
Copy a string without overflow (or indication of it :-).
*/ 

int strzcpy
(
char *String,
char *cptr,
int SizeOfString
)
{
   char  *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (SizeOfString) SizeOfString--;
   zptr = (sptr = String) + SizeOfString;
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';
   return (sptr - String);
}

/****************************************************************************/
/*
Does a case-insensitive, character-by-character string compare and returns 
TRUE if two strings are the same, or FALSE if not.  If a maximum number of 
characters are specified only those will be compared, if the entire strings 
should be compared then specify the number of characters as 0.
*/ 

BOOL strsame
(
char *sptr1,
char *sptr2,
int  count
)
{
   /*********/
   /* begin */
   /*********/

   while (*sptr1 && *sptr2)
   {
      if (toupper (*sptr1++) != toupper (*sptr2++)) return (FALSE);
      if (count)
         if (!--count) return (TRUE);
   }
   if (*sptr1 || *sptr2)
      return (FALSE);
   else
      return (TRUE);
}

/*****************************************************************************/

